import { Connection } from "@grapecity-software/js-collaboration-client";

/**
 * Error codes of ot errors.
 */
export enum ERROR_CODES {
    ERR_DOC_ALREADY_CREATED,
    ERR_DOC_DOES_NOT_EXIST,
    ERR_DOC_TYPE_NOT_RECOGNIZED,
    ERR_DOC_WAS_DELETED,
    ERR_MAX_SUBMIT_RETRIES_EXCEEDED,
    ERR_MAX_SUBMIT_SNAPSHOT_RETRIES_EXCEEDED,
    ERR_MESSAGE_BADLY_FORMED,
    ERR_OP_ALREADY_SUBMITTED,
    ERR_HARD_ROLLBACK_FETCH_FAILED,
    ERR_OT_OP_NOT_APPLIED,
    ERR_OT_OP_TRANSFORM_FAILED,
    /**
     * A "readSnapshots" middleware rejected the reads of specific snapshots.
     *
     * This error code is mostly for server use and generally will not be encountered on the client.
     * Instead, each specific doc that encountered an error will receive its specific error.
     *
     * The one exception is for queries, where a "readSnapshots" rejection of specific snapshots will
     * cause the client to receive this error for the whole query, since queries don't support
     * doc-specific errors.
     */
    ERR_SNAPSHOT_READS_REJECTED,
    ERR_SUBMIT_TRANSFORM_OPS_NOT_FOUND,
    ERR_UNKNOWN_ERROR,
    ERR_HISTORY_SNAPSHOT_NOT_FOUND,
}

export interface IIdToError {
    [docId: string]: string | OTError;
}

export declare class OTError extends Error {
    static CODES: Record<string, string>;
    name: string;
    data?: unknown;
    idToError?: IIdToError | null;
}

export interface IOp<T = unknown> {
    v: number;
    create?: ICreateComponent;
    op?: T;
    del?: boolean;
}

export interface ICreateComponent {
    type: string;
    data: unknown;
}

export interface ISnapshot<S = unknown> {
    id: string;
    v: number;
    type?: string;
    data?: S;
    /**
     * only server side has this field
     */
    fragments?: ISnapshotFragments;
}

export type ISnapshotFragments<S = unknown> = { [key: string]: S };

export interface ISnapshotMeta {
    mtime?: number;
    ctime?: number;
}

export interface IRequestMessage<T> {
    /**
     * message action
     */
    a: MessageActions;
    /**
     * version
     */
    v?: number;
    create?: ICreateComponent;
    op?: T;
    del?: boolean;
    error?: IErrorData;
}

export interface IErrorData {
    code: number;
    message: string;
}

export interface IReplyMessage<S, T> {
    a: MessageActions;
    v?: number;
    create?: ICreateComponent;
    op?: T;
    del?: boolean;
    error?: IErrorData;
    data?: ISnapshot<S>;
}

/**
 * Message actions.
 */
export enum MessageActions {
    fetch,
    fetchOps,
    fetchHistorySnapshot,
    subscribe,
    unsubscribe,
    op,
}

export interface ICallBack<T = unknown> {
    (err?: Error | null, data?: T): void;
}

/**
 * Events emitted by the SharedDoc.
 * @template S - The type of snapshot data.
 * @template T - The type of operation data.
 */
export interface ISharedDocEvents<S, T> {
    /**
     * Triggered when an error occurs.
     */
    error: (err: OTError) => void;
    /**
     * Triggered after a client fetches or subscribes to the snapshot.
     */
    load: () => void;
    create: (source?: unknown) => void;
    del: (data: S, source?: unknown) => void;
    hardRollback: () => void;
    /**
     * Triggered before an operation is applied.
     */
    beforeOp: (op: T, source?: unknown) => void;
    /**
     * Triggered after an operation is applied.
     */
    op: (op: T, source?: unknown) => void;
}

/**
 * Represents a shared document for requesting snapshots and operations.
 * @template S - The type of snapshot data.
 * @template T - The type of operation data.
 */
export declare class SharedDoc<S = unknown, T = unknown> {
    /**
     * Creates a new shared document instance.
     */
    constructor(conn: Connection);

    /**
     * Retrieves the ID of the shared document.
     * @returns {string} The document ID.
     */
    get id(): string;
    /**
     * Retrieves the connection associated with the shared document.
     * @returns {Connection} The connection instance.
     */
    get connection(): Connection;
    /**
     * Retrieves the version of the current snapshot.
     */
    get version(): number | undefined;
    /**
     * Retrieves the OT type of the snapshot.
     */
    get type(): OT_Type<S, T> | undefined;
    /**
     * Retrieves the data of the current snapshot.
     */
    get data(): S | undefined;
    /**
     * Registers a listener for a specific event.
     * @template NAME - The type of the event name, extending keyof ISharedDocEvents<S, T>.
     * @param {NAME} name - The name of the event.
     * @param {ISharedDocEvents<S, T>[NAME]} f - The event handler function.
     * @returns {ISharedDocEvents<S, T>[NAME]} The registered event handler.
     */
    on<NAME extends keyof ISharedDocEvents<S, T>>(name: NAME, f: ISharedDocEvents<S, T>[NAME]): ISharedDocEvents<S, T>[NAME];
    /**
     * Registers a one-time listener for a specific event.
     * @template NAME - The type of the event name, extending keyof ISharedDocEvents<S, T>.
     * @param {NAME} name - The name of the event.
     * @param {ISharedDocEvents<S, T>[NAME]} f - The event handler function to call once.
     */
    once<NAME extends keyof ISharedDocEvents<S, T>>(name: NAME, f: ISharedDocEvents<S, T>[NAME]): void;
    /**
     * Removes a listener for a specific event.
     * @template NAME - The type of the event name, extending keyof ISharedDocEvents<S, T>.
     * @param {NAME} name - The name of the event.
     * @param {ISharedDocEvents<S, T>[NAME]} f - The event handler function to remove.
     */
    off<NAME extends keyof ISharedDocEvents<S, T>>(name: NAME, f: ISharedDocEvents<S, T>[NAME]): void;
    /**
     * Destroys the observable and cleans up resources.
     */
    destroy(): void;

    /**
     * Fetches the snapshot data.
     * @returns {Promise<void>} A promise that resolves after the snapshot is fetched, the fetched snapshot data is 'doc.data'.
     */
    fetch(): Promise<void>;
    /**
     * Subscribes the snapshot data.
     * @returns {Promise<void>} A promise that resolves after the snapshot is subscribed, the subscribed snapshot data is 'doc.data'.
     */
    subscribe(): Promise<void>;
    /**
     * Submits an op to the document.
     * @template T - The type of op component.
     * @param {T} component - The op component.
     * @param {IOptions} [options] - Options for the operation (optional).
     * @returns {Promise<void>} A promise that resolves after the op is submitted.
     */
    submitOp(component: T, options?: IOptions): Promise<void>;
    /**
     * Create a new document.
     * @template S - The type of initialize snapshot data.
     * @param {S} data - The data used to initialize the snapshot.
     * @param {string} type - The URI of the OT type.
     * @param {IOptions} options - Options for the operation.
     * @returns {Promise<void>} A promise that resolves after the document is created.
     */
    create(data: S, type: string, options: IOptions): Promise<void>;
    /**
     * Fetches the history snapshot data.
     * @param {IHistorySnapshotRequestOptions} options - Options for the history snapshot request.
     * @returns {Promise<ISnapshot<S>>} A promise that resolves after the history snapshot is fetched.
     */
    fetchHistorySnapshot(options: IHistorySnapshotRequestOptions): Promise<ISnapshot<S>>;
    /**
     * Delete the document.
     * @param {IOptions} options - The options of the op.
     * @returns {Promise<void>} A promise that resolves after the document is deleted.
     */
    del(options: IOptions): Promise<void>;
    /**
     * Hard rolls back the snapshot data to the latest version.
     * @returns {Promise<void>} A promise that resolves when the rollback is complete.
     */
    hardRollback(): Promise<void>;
}

/**
 * Defines the interface for customizing operational transformation (OT) behavior in collaboration.
 * @template S The type of snapshot data.
 * @template T The type of op data.
 */
export interface OT_Type<S = unknown, T = unknown> {
    /**
     * The URI identifying the document type.
     */
    uri: string;
    /**
     * Creates initial snapshot data from provided data.
     * @param {S} data - The initial snapshot data.
     * @returns {S} The created snapshot data.
     */
    create?(data: S): S;
    /**
     * Transforms an operation to resolve conflicts with another operation.
     * @param {T} op1 - The operation to transform.
     * @param {T} op2 - The operation serving as the basis for transformation.
     * @param {'left' | 'right'} side - Indicates which operation arrived at the server later.
     * @returns {T} The transformed operation.
     */
    transform(op1: T, op2: T, side: 'left' | 'right'): T;
    /**
     * Transforms two operations against each other.
     * @param op1 The op to be transformed based on op2.
     * @param op2 The op to be transformed based on op1.
     * @returns An array containing the transformed op1 and op2.
     */
    transformX?(op1: T, op2: T): T[];
    /**
     * Applies an operation to snapshot data.
     * @param {S} snapshot - The current snapshot data.
     * @param {T} op - The operation to apply.
     * @returns {S} The updated snapshot data.
     */
    apply?(snapshot: S, op: T): S;
    /**
     * Normalizes an operation.
     * @param {T} op - The operation to normalize.
     * @returns {T} The normalized operation.
     */
    normalize?(op: T): T;
    /**
     * Composes two operations into a single operation.
     * @param {T} op1 - The first operation to compose.
     * @param {T} op2 - The second operation to compose with op1.
     * @returns {T} The composed operation.
     */
    compose?(op1: T, op2: T): T;
    /**
     * Deserializes snapshot data for client use.
     * @param {S} snapshot - The snapshot data to deserialize.
     * @returns {S} The deserialized snapshot data.
     */
    deserialize?(snapshot: S): S;
    /**
     * Creates deserialized client document data from snapshot data.
     * @param {S} data - The snapshot data to deserialize.
     * @returns {S | undefined} The deserialized snapshot data, or undefined if not applicable.
     */
    createDeserialized?(data: S): S | undefined;
}

/**
 * Manages operational transformation (OT) types for collaboration, enabling registration and retrieval by URI.
 */
export declare class TypesManager {
    /**
     * Registers a custom OT type.
     * @template S - The type of snapshot data.
     * @template T - The type of operation data.
     * @param {OT_Type<S, T>} type - The user-defined OT type to register.
     */
    static register<S, T>(type: OT_Type<S, T>): void;
    /**
     * Retrieves an OT type by its URI.
     * @template S - The type of snapshot data.
     * @template T - The type of operation data.
     * @param {string} typeUri - The URI of the type to retrieve.
     * @returns {OT_Type<S, T>} The OT type associated with the specified URI.
     */
    static getType<S, T>(typeUri: string): OT_Type<S, T>;
}

export interface IHistorySnapshotRequestOptions {
    version?: number;
}

export interface IOptions {
    source?: string;
}
